# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .all_axes import AllAxes as AllAxes
from .axis import Axis as Axis
from .axis_group import AxisGroup as AxisGroup
from .axis_settings import AxisSettings as AxisSettings
from .connection import Connection as Connection
from .device import Device as Device
from .device_io import DeviceIO as DeviceIO
from .device_settings import DeviceSettings as DeviceSettings
from .lockstep import Lockstep as Lockstep
from .oscilloscope import Oscilloscope as Oscilloscope
from .oscilloscope_data import OscilloscopeData as OscilloscopeData
from .pvt import Pvt as Pvt
from .pvt_buffer import PvtBuffer as PvtBuffer
from .pvt_io import PvtIo as PvtIo
from .pvt_sequence import PvtSequence as PvtSequence
from .servo_tuner import ServoTuner as ServoTuner
from .setting_constants import SettingConstants as SettingConstants
from .storage import AxisStorage as AxisStorage, DeviceStorage as DeviceStorage
from .stream import Stream as Stream
from .stream_buffer import StreamBuffer as StreamBuffer
from .stream_io import StreamIo as StreamIo
from .streams import Streams as Streams
from .transport import Transport as Transport
from .trigger import Trigger as Trigger
from .triggers import Triggers as Triggers
from .warning_flags import WarningFlags as WarningFlags
from .warnings import Warnings as Warnings
from ..dto.ascii.alert_event import AlertEvent as AlertEvent
from ..dto.ascii.axis_identity import AxisIdentity as AxisIdentity
from ..dto.ascii.axis_type import AxisType as AxisType
from ..dto.ascii.can_set_state_axis_response import CanSetStateAxisResponse as CanSetStateAxisResponse
from ..dto.ascii.can_set_state_device_response import CanSetStateDeviceResponse as CanSetStateDeviceResponse
from ..dto.ascii.conversion_factor import ConversionFactor as ConversionFactor
from ..dto.ascii.device_identity import DeviceIdentity as DeviceIdentity
from ..dto.ascii.device_io_info import DeviceIOInfo as DeviceIOInfo
from ..dto.ascii.digital_output_action import DigitalOutputAction as DigitalOutputAction
from ..dto.ascii.get_axis_setting import GetAxisSetting as GetAxisSetting
from ..dto.ascii.get_axis_setting_result import GetAxisSettingResult as GetAxisSettingResult
from ..dto.ascii.get_setting import GetSetting as GetSetting
from ..dto.ascii.get_setting_result import GetSettingResult as GetSettingResult
from ..dto.ascii.io_port_label import IoPortLabel as IoPortLabel
from ..dto.ascii.io_port_type import IoPortType as IoPortType
from ..dto.ascii.lockstep_axes import LockstepAxes as LockstepAxes
from ..dto.ascii.message_type import MessageType as MessageType
from ..dto.ascii.oscilloscope_capture_properties import OscilloscopeCaptureProperties as OscilloscopeCaptureProperties
from ..dto.ascii.oscilloscope_data_source import OscilloscopeDataSource as OscilloscopeDataSource
from ..dto.ascii.paramset_info import ParamsetInfo as ParamsetInfo
from ..dto.ascii.pid_tuning import PidTuning as PidTuning
from ..dto.ascii.pvt_axis_definition import PvtAxisDefinition as PvtAxisDefinition
from ..dto.ascii.pvt_axis_type import PvtAxisType as PvtAxisType
from ..dto.ascii.pvt_mode import PvtMode as PvtMode
from ..dto.ascii.response import Response as Response
from ..dto.ascii.servo_tuning_param import ServoTuningParam as ServoTuningParam
from ..dto.ascii.servo_tuning_paramset import ServoTuningParamset as ServoTuningParamset
from ..dto.ascii.simple_tuning import SimpleTuning as SimpleTuning
from ..dto.ascii.simple_tuning_param_definition import SimpleTuningParamDefinition as SimpleTuningParamDefinition
from ..dto.ascii.stream_axis_definition import StreamAxisDefinition as StreamAxisDefinition
from ..dto.ascii.stream_axis_type import StreamAxisType as StreamAxisType
from ..dto.ascii.stream_mode import StreamMode as StreamMode
from ..dto.ascii.trigger_action import TriggerAction as TriggerAction
from ..dto.ascii.trigger_condition import TriggerCondition as TriggerCondition
from ..dto.ascii.trigger_enabled_state import TriggerEnabledState as TriggerEnabledState
from ..dto.ascii.trigger_operation import TriggerOperation as TriggerOperation
from ..dto.ascii.trigger_state import TriggerState as TriggerState
from ..dto.ascii.unknown_response_event import UnknownResponseEvent as UnknownResponseEvent



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import zaber_bson
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class GetAxisSetting:
    """
    Specifies a setting to get with one of the multi-get commands.
    """

    setting: str
    """
    The setting to read.
    """

    unit: Optional[UnitsAndLiterals] = None
    """
    The unit to convert the read setting to.
    """

    @staticmethod
    def zero_values() -> 'GetAxisSetting':
        return GetAxisSetting(
            setting="",
            unit=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'GetAxisSetting':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return GetAxisSetting.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'setting': str(self.setting or ''),
            'unit': units_from_literals(self.unit).value if self.unit is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'GetAxisSetting':
        return GetAxisSetting(
            setting=data.get('setting'),  # type: ignore
            unit=Units(data.get('unit')) if data.get('unit') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.setting is not None:
            if not isinstance(self.setting, str):
                raise ValueError(f'Property "Setting" of "GetAxisSetting" is not a string.')

        if self.unit is not None:
            if not isinstance(self.unit, (Units, str)):
                raise ValueError(f'Property "Unit" of "GetAxisSetting" is not Units.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from ..binary.command_code import CommandCode
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class BinaryGenericWithUnitsRequest:

    interface_id: int = 0

    device: int = 0

    command: CommandCode = next(first for first in CommandCode)

    data: float = 0

    from_unit: UnitsAndLiterals = Units.NATIVE

    to_unit: UnitsAndLiterals = Units.NATIVE

    timeout: float = 0

    @staticmethod
    def zero_values() -> 'BinaryGenericWithUnitsRequest':
        return BinaryGenericWithUnitsRequest(
            interface_id=0,
            device=0,
            command=next(first for first in CommandCode),
            data=0,
            from_unit=Units.NATIVE,
            to_unit=Units.NATIVE,
            timeout=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'BinaryGenericWithUnitsRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return BinaryGenericWithUnitsRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'command': self.command.value,
            'data': float(self.data),
            'fromUnit': units_from_literals(self.from_unit).value,
            'toUnit': units_from_literals(self.to_unit).value,
            'timeout': float(self.timeout),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'BinaryGenericWithUnitsRequest':
        return BinaryGenericWithUnitsRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            command=CommandCode(data.get('command')),  # type: ignore
            data=data.get('data'),  # type: ignore
            from_unit=Units(data.get('fromUnit')),  # type: ignore
            to_unit=Units(data.get('toUnit')),  # type: ignore
            timeout=data.get('timeout'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "BinaryGenericWithUnitsRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "BinaryGenericWithUnitsRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "BinaryGenericWithUnitsRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "BinaryGenericWithUnitsRequest" is not integer value.')

        if self.command is None:
            raise ValueError(f'Property "Command" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.command, CommandCode):
            raise ValueError(f'Property "Command" of "BinaryGenericWithUnitsRequest" is not an instance of "CommandCode".')

        if self.data is None:
            raise ValueError(f'Property "Data" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.data, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Data" of "BinaryGenericWithUnitsRequest" is not a number.')

        if self.from_unit is None:
            raise ValueError(f'Property "FromUnit" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.from_unit, (Units, str)):
            raise ValueError(f'Property "FromUnit" of "BinaryGenericWithUnitsRequest" is not Units.')

        if self.to_unit is None:
            raise ValueError(f'Property "ToUnit" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.to_unit, (Units, str)):
            raise ValueError(f'Property "ToUnit" of "BinaryGenericWithUnitsRequest" is not Units.')

        if self.timeout is None:
            raise ValueError(f'Property "Timeout" of "BinaryGenericWithUnitsRequest" is None.')

        if not isinstance(self.timeout, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Timeout" of "BinaryGenericWithUnitsRequest" is not a number.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class CommandFailedExceptionData:
    """
    Contains additional data for CommandFailedException.
    """

    command: str
    """
    The command that got rejected.
    """

    response_data: str
    """
    The data from the reply containing the rejection reason.
    """

    reply_flag: str
    """
    The flag indicating that the command was rejected.
    """

    status: str
    """
    The current device or axis status.
    """

    warning_flag: str
    """
    The highest priority warning flag on the device or axis.
    """

    device_address: int
    """
    The address of the device that rejected the command.
    """

    axis_number: int
    """
    The number of the axis which the rejection relates to.
    """

    id: int
    """
    The message ID of the reply.
    """

    @staticmethod
    def zero_values() -> 'CommandFailedExceptionData':
        return CommandFailedExceptionData(
            command="",
            response_data="",
            reply_flag="",
            status="",
            warning_flag="",
            device_address=0,
            axis_number=0,
            id=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'CommandFailedExceptionData':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return CommandFailedExceptionData.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'command': str(self.command or ''),
            'responseData': str(self.response_data or ''),
            'replyFlag': str(self.reply_flag or ''),
            'status': str(self.status or ''),
            'warningFlag': str(self.warning_flag or ''),
            'deviceAddress': int(self.device_address),
            'axisNumber': int(self.axis_number),
            'id': int(self.id),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'CommandFailedExceptionData':
        return CommandFailedExceptionData(
            command=data.get('command'),  # type: ignore
            response_data=data.get('responseData'),  # type: ignore
            reply_flag=data.get('replyFlag'),  # type: ignore
            status=data.get('status'),  # type: ignore
            warning_flag=data.get('warningFlag'),  # type: ignore
            device_address=data.get('deviceAddress'),  # type: ignore
            axis_number=data.get('axisNumber'),  # type: ignore
            id=data.get('id'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.command is not None:
            if not isinstance(self.command, str):
                raise ValueError(f'Property "Command" of "CommandFailedExceptionData" is not a string.')

        if self.response_data is not None:
            if not isinstance(self.response_data, str):
                raise ValueError(f'Property "ResponseData" of "CommandFailedExceptionData" is not a string.')

        if self.reply_flag is not None:
            if not isinstance(self.reply_flag, str):
                raise ValueError(f'Property "ReplyFlag" of "CommandFailedExceptionData" is not a string.')

        if self.status is not None:
            if not isinstance(self.status, str):
                raise ValueError(f'Property "Status" of "CommandFailedExceptionData" is not a string.')

        if self.warning_flag is not None:
            if not isinstance(self.warning_flag, str):
                raise ValueError(f'Property "WarningFlag" of "CommandFailedExceptionData" is not a string.')

        if self.device_address is None:
            raise ValueError(f'Property "DeviceAddress" of "CommandFailedExceptionData" is None.')

        if not isinstance(self.device_address, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "DeviceAddress" of "CommandFailedExceptionData" is not a number.')

        if int(self.device_address) != self.device_address:
            raise ValueError(f'Property "DeviceAddress" of "CommandFailedExceptionData" is not integer value.')

        if self.axis_number is None:
            raise ValueError(f'Property "AxisNumber" of "CommandFailedExceptionData" is None.')

        if not isinstance(self.axis_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisNumber" of "CommandFailedExceptionData" is not a number.')

        if int(self.axis_number) != self.axis_number:
            raise ValueError(f'Property "AxisNumber" of "CommandFailedExceptionData" is not integer value.')

        if self.id is None:
            raise ValueError(f'Property "Id" of "CommandFailedExceptionData" is None.')

        if not isinstance(self.id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Id" of "CommandFailedExceptionData" is not a number.')

        if int(self.id) != self.id:
            raise ValueError(f'Property "Id" of "CommandFailedExceptionData" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from .process_controller_source_sensor import ProcessControllerSourceSensor


@dataclass
class ProcessControllerSource:
    """
    The source used by a process in a closed-loop mode.
    """

    sensor: ProcessControllerSourceSensor
    """
    The type of input sensor.
    """

    port: int
    """
    The specific input to use.
    """

    @staticmethod
    def zero_values() -> 'ProcessControllerSource':
        return ProcessControllerSource(
            sensor=next(first for first in ProcessControllerSourceSensor),
            port=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'ProcessControllerSource':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return ProcessControllerSource.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'sensor': self.sensor.value,
            'port': int(self.port),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'ProcessControllerSource':
        return ProcessControllerSource(
            sensor=ProcessControllerSourceSensor(data.get('sensor')),  # type: ignore
            port=data.get('port'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.sensor is None:
            raise ValueError(f'Property "Sensor" of "ProcessControllerSource" is None.')

        if not isinstance(self.sensor, ProcessControllerSourceSensor):
            raise ValueError(f'Property "Sensor" of "ProcessControllerSource" is not an instance of "ProcessControllerSourceSensor".')

        if self.port is None:
            raise ValueError(f'Property "Port" of "ProcessControllerSource" is None.')

        if not isinstance(self.port, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Port" of "ProcessControllerSource" is not a number.')

        if int(self.port) != self.port:
            raise ValueError(f'Property "Port" of "ProcessControllerSource" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class FirmwareVersion:
    """
    Class representing version of firmware in the controller.
    """

    major: int
    """
    Major version number.
    """

    minor: int
    """
    Minor version number.
    """

    build: int
    """
    Build version number.
    """

    @staticmethod
    def zero_values() -> 'FirmwareVersion':
        return FirmwareVersion(
            major=0,
            minor=0,
            build=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'FirmwareVersion':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return FirmwareVersion.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'major': int(self.major),
            'minor': int(self.minor),
            'build': int(self.build),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'FirmwareVersion':
        return FirmwareVersion(
            major=data.get('major'),  # type: ignore
            minor=data.get('minor'),  # type: ignore
            build=data.get('build'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.major is None:
            raise ValueError(f'Property "Major" of "FirmwareVersion" is None.')

        if not isinstance(self.major, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Major" of "FirmwareVersion" is not a number.')

        if int(self.major) != self.major:
            raise ValueError(f'Property "Major" of "FirmwareVersion" is not integer value.')

        if self.minor is None:
            raise ValueError(f'Property "Minor" of "FirmwareVersion" is None.')

        if not isinstance(self.minor, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Minor" of "FirmwareVersion" is not a number.')

        if int(self.minor) != self.minor:
            raise ValueError(f'Property "Minor" of "FirmwareVersion" is not integer value.')

        if self.build is None:
            raise ValueError(f'Property "Build" of "FirmwareVersion" is None.')

        if not isinstance(self.build, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Build" of "FirmwareVersion" is not a number.')

        if int(self.build) != self.build:
            raise ValueError(f'Property "Build" of "FirmwareVersion" is not integer value.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
# pylint: disable=W0622

from ..call import call, call_async, call_sync
from ..dto import requests as dto
from ..ascii import Device
from ..units import TimeUnits, Units


class CameraTrigger:
    """
    An abstraction over a device and it's digital output channel.
    """

    @property
    def device(self) -> Device:
        """
        The device whose digital output triggers the camera.
        """
        return self._device

    @property
    def channel(self) -> int:
        """
        The digital output channel that triggers the camera.
        """
        return self._channel

    def __init__(self, device: Device, channel: int):
        """
        Creates instance of `CameraTrigger` based on the given device and digital output channel.
        """
        self._device = device
        self._channel = channel

    def trigger(
            self,
            pulse_width: float,
            unit: TimeUnits = Units.NATIVE,
            wait: bool = True
    ) -> None:
        """
        Triggers the camera.
        Schedules trigger pulse on the digital output channel.
        By default, the method waits until the trigger pulse is finished.

        Args:
            pulse_width: The time duration of the trigger pulse.
                Depending on the camera setting, the argument can be use to specify exposure.
            unit: Units of time.
            wait: If false, the method does not wait until the trigger pulse is finished.
        """
        request = dto.MicroscopeTriggerCameraRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            channel_number=self.channel,
            delay=pulse_width,
            unit=unit,
            wait=wait,
        )
        call("microscope/trigger_camera", request)

    async def trigger_async(
            self,
            pulse_width: float,
            unit: TimeUnits = Units.NATIVE,
            wait: bool = True
    ) -> None:
        """
        Triggers the camera.
        Schedules trigger pulse on the digital output channel.
        By default, the method waits until the trigger pulse is finished.

        Args:
            pulse_width: The time duration of the trigger pulse.
                Depending on the camera setting, the argument can be use to specify exposure.
            unit: Units of time.
            wait: If false, the method does not wait until the trigger pulse is finished.
        """
        request = dto.MicroscopeTriggerCameraRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            channel_number=self.channel,
            delay=pulse_width,
            unit=unit,
            wait=wait,
        )
        await call_async("microscope/trigger_camera", request)

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call_sync(
            "device/device_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# This file is generated. Do not modify by hand.
from enum import Enum


class AxisType(Enum):
    """
    Denotes type of an axis and units it accepts.
    """

    UNKNOWN = 0
    LINEAR = 1
    ROTARY = 2
    PROCESS = 3
    LAMP = 4



import struct
from ctypes import string_at, c_void_p
from functools import reduce
from typing import List

SIZE_TYPE_SIZE = 4
SIZE_TYPE_PACKING = "I"


def serialize(buffers: List[bytes]) -> bytes:
    size = SIZE_TYPE_SIZE + reduce((lambda sum, buffer: sum + len(buffer) + SIZE_TYPE_SIZE), buffers, 0)

    buffer = struct.pack(SIZE_TYPE_PACKING, size)

    for messageBuffer in buffers:
        buffer = buffer + struct.pack(SIZE_TYPE_PACKING, len(messageBuffer)) + bytes(messageBuffer)

    return buffer


def deserialize(data: c_void_p) -> List[bytes]:
    total_size = struct.unpack_from(SIZE_TYPE_PACKING, string_at(data, SIZE_TYPE_SIZE))[0]
    offset = SIZE_TYPE_SIZE

    buffers = []
    while offset < total_size:
        size = struct.unpack_from(SIZE_TYPE_PACKING, string_at(data + offset, SIZE_TYPE_SIZE))[0]  # type: ignore
        offset += SIZE_TYPE_SIZE
        buffers.append(string_at(data + offset, size))  # type: ignore
        offset += size

    return buffers



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class PidTuning:
    """
    The tuning of this axis represented by PID parameters.
    """

    type: str
    """
    The tuning algorithm used to tune this axis.
    """

    version: int
    """
    The version of the tuning algorithm used to tune this axis.
    """

    p: float
    """
    The positional tuning argument.
    """

    i: float
    """
    The integral tuning argument.
    """

    d: float
    """
    The derivative tuning argument.
    """

    fc: float
    """
    The frequency cutoff for the tuning.
    """

    @staticmethod
    def zero_values() -> 'PidTuning':
        return PidTuning(
            type="",
            version=0,
            p=0,
            i=0,
            d=0,
            fc=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'PidTuning':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return PidTuning.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'type': str(self.type or ''),
            'version': int(self.version),
            'p': float(self.p),
            'i': float(self.i),
            'd': float(self.d),
            'fc': float(self.fc),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'PidTuning':
        return PidTuning(
            type=data.get('type'),  # type: ignore
            version=data.get('version'),  # type: ignore
            p=data.get('p'),  # type: ignore
            i=data.get('i'),  # type: ignore
            d=data.get('d'),  # type: ignore
            fc=data.get('fc'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.type is not None:
            if not isinstance(self.type, str):
                raise ValueError(f'Property "Type" of "PidTuning" is not a string.')

        if self.version is None:
            raise ValueError(f'Property "Version" of "PidTuning" is None.')

        if not isinstance(self.version, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Version" of "PidTuning" is not a number.')

        if int(self.version) != self.version:
            raise ValueError(f'Property "Version" of "PidTuning" is not integer value.')

        if self.p is None:
            raise ValueError(f'Property "P" of "PidTuning" is None.')

        if not isinstance(self.p, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "P" of "PidTuning" is not a number.')

        if self.i is None:
            raise ValueError(f'Property "I" of "PidTuning" is None.')

        if not isinstance(self.i, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "I" of "PidTuning" is not a number.')

        if self.d is None:
            raise ValueError(f'Property "D" of "PidTuning" is None.')

        if not isinstance(self.d, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "D" of "PidTuning" is not a number.')

        if self.fc is None:
            raise ValueError(f'Property "Fc" of "PidTuning" is None.')

        if not isinstance(self.fc, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Fc" of "PidTuning" is not a number.')



# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .offline_translator import OfflineTranslator as OfflineTranslator
from .translator import Translator as Translator
from ..dto.gcode.axis_definition import AxisDefinition as AxisDefinition
from ..dto.gcode.axis_mapping import AxisMapping as AxisMapping
from ..dto.gcode.axis_transformation import AxisTransformation as AxisTransformation
from ..dto.gcode.device_definition import DeviceDefinition as DeviceDefinition
from ..dto.gcode.translate_message import TranslateMessage as TranslateMessage
from ..dto.gcode.translate_result import TranslateResult as TranslateResult
from ..dto.gcode.translator_config import TranslatorConfig as TranslatorConfig



# This file is generated. Do not modify by hand.
from enum import Enum


class TriggerOperation(Enum):
    """
    Operation for trigger action.
    """

    SET_TO = 0
    INCREMENT_BY = 1
    DECREMENT_BY = 2



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import zaber_bson


@dataclass
class GatewayRequest:

    request: str = ""

    @staticmethod
    def zero_values() -> 'GatewayRequest':
        return GatewayRequest(
            request="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'GatewayRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return GatewayRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'request': str(self.request or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'GatewayRequest':
        return GatewayRequest(
            request=data.get('request'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.request is not None:
            if not isinstance(self.request, str):
                raise ValueError(f'Property "Request" of "GatewayRequest" is not a string.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .command_failed_exception import CommandFailedException


class RemoteModeException(CommandFailedException):
    """
    Thrown when a command is rejected because the device is in EtherCAT Control (remote) mode.
    """



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional
import decimal
from collections.abc import Iterable
import zaber_bson
from ..ascii.servo_tuning_paramset import ServoTuningParamset
from ..ascii.servo_tuning_param import ServoTuningParam


@dataclass
class SetSimpleTuning:

    interface_id: int = 0

    device: int = 0

    axis: int = 0

    paramset: ServoTuningParamset = next(first for first in ServoTuningParamset)

    load_mass: float = 0

    tuning_params: List[ServoTuningParam] = field(default_factory=list)

    carriage_mass: Optional[float] = None

    @staticmethod
    def zero_values() -> 'SetSimpleTuning':
        return SetSimpleTuning(
            interface_id=0,
            device=0,
            axis=0,
            paramset=next(first for first in ServoTuningParamset),
            carriage_mass=None,
            load_mass=0,
            tuning_params=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SetSimpleTuning':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SetSimpleTuning.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'axis': int(self.axis),
            'paramset': self.paramset.value,
            'carriageMass': float(self.carriage_mass) if self.carriage_mass is not None else None,
            'loadMass': float(self.load_mass),
            'tuningParams': [item.to_dict() for item in self.tuning_params] if self.tuning_params is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SetSimpleTuning':
        return SetSimpleTuning(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            axis=data.get('axis'),  # type: ignore
            paramset=ServoTuningParamset(data.get('paramset')),  # type: ignore
            carriage_mass=data.get('carriageMass'),  # type: ignore
            load_mass=data.get('loadMass'),  # type: ignore
            tuning_params=[ServoTuningParam.from_dict(item) for item in data.get('tuningParams')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "SetSimpleTuning" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "SetSimpleTuning" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "SetSimpleTuning" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "SetSimpleTuning" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "SetSimpleTuning" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "SetSimpleTuning" is not integer value.')

        if self.axis is None:
            raise ValueError(f'Property "Axis" of "SetSimpleTuning" is None.')

        if not isinstance(self.axis, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis" of "SetSimpleTuning" is not a number.')

        if int(self.axis) != self.axis:
            raise ValueError(f'Property "Axis" of "SetSimpleTuning" is not integer value.')

        if self.paramset is None:
            raise ValueError(f'Property "Paramset" of "SetSimpleTuning" is None.')

        if not isinstance(self.paramset, ServoTuningParamset):
            raise ValueError(f'Property "Paramset" of "SetSimpleTuning" is not an instance of "ServoTuningParamset".')

        if self.carriage_mass is not None:
            if not isinstance(self.carriage_mass, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "CarriageMass" of "SetSimpleTuning" is not a number.')

        if self.load_mass is None:
            raise ValueError(f'Property "LoadMass" of "SetSimpleTuning" is None.')

        if not isinstance(self.load_mass, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "LoadMass" of "SetSimpleTuning" is not a number.')

        if self.tuning_params is not None:
            if not isinstance(self.tuning_params, Iterable):
                raise ValueError('Property "TuningParams" of "SetSimpleTuning" is not iterable.')

            for i, tuning_params_item in enumerate(self.tuning_params):
                if tuning_params_item is None:
                    raise ValueError(f'Item {i} in property "TuningParams" of "SetSimpleTuning" is None.')

                if not isinstance(tuning_params_item, ServoTuningParam):
                    raise ValueError(f'Item {i} in property "TuningParams" of "SetSimpleTuning" is not an instance of "ServoTuningParam".')

                tuning_params_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
from collections.abc import Iterable
import zaber_bson
from ..ascii.response import Response


@dataclass
class GenericCommandResponseCollection:

    responses: List[Response] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'GenericCommandResponseCollection':
        return GenericCommandResponseCollection(
            responses=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'GenericCommandResponseCollection':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return GenericCommandResponseCollection.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'responses': [item.to_dict() for item in self.responses] if self.responses is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'GenericCommandResponseCollection':
        return GenericCommandResponseCollection(
            responses=[Response.from_dict(item) for item in data.get('responses')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.responses is not None:
            if not isinstance(self.responses, Iterable):
                raise ValueError('Property "Responses" of "GenericCommandResponseCollection" is not iterable.')

            for i, responses_item in enumerate(self.responses):
                if responses_item is None:
                    raise ValueError(f'Item {i} in property "Responses" of "GenericCommandResponseCollection" is None.')

                if not isinstance(responses_item, Response):
                    raise ValueError(f'Item {i} in property "Responses" of "GenericCommandResponseCollection" is not an instance of "Response".')

                responses_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
from collections.abc import Iterable
import zaber_bson


@dataclass
class StreamMovementInterruptedExceptionData:
    """
    Contains additional data for StreamMovementInterruptedException.
    """

    warnings: List[str]
    """
    The full list of warnings.
    """

    reason: str
    """
    The reason for the Exception.
    """

    @staticmethod
    def zero_values() -> 'StreamMovementInterruptedExceptionData':
        return StreamMovementInterruptedExceptionData(
            warnings=[],
            reason="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'StreamMovementInterruptedExceptionData':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return StreamMovementInterruptedExceptionData.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'warnings': [str(item or '') for item in self.warnings] if self.warnings is not None else [],
            'reason': str(self.reason or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'StreamMovementInterruptedExceptionData':
        return StreamMovementInterruptedExceptionData(
            warnings=data.get('warnings'),  # type: ignore
            reason=data.get('reason'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.warnings is not None:
            if not isinstance(self.warnings, Iterable):
                raise ValueError('Property "Warnings" of "StreamMovementInterruptedExceptionData" is not iterable.')

            for i, warnings_item in enumerate(self.warnings):
                if warnings_item is not None:
                    if not isinstance(warnings_item, str):
                        raise ValueError(f'Item {i} in property "Warnings" of "StreamMovementInterruptedExceptionData" is not a string.')

        if self.reason is not None:
            if not isinstance(self.reason, str):
                raise ValueError(f'Property "Reason" of "StreamMovementInterruptedExceptionData" is not a string.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, Optional
from ..dto import requests as dto
from ..call import call, call_async
from ..units import Units, LengthUnits, UnitsAndLiterals, TimeUnits
from ..dto.ascii.trigger_state import TriggerState
from ..dto.ascii.trigger_enabled_state import TriggerEnabledState
from ..dto.ascii.io_port_type import IoPortType
from ..dto.ascii.trigger_condition import TriggerCondition
from ..dto.ascii.trigger_action import TriggerAction
from ..dto.ascii.trigger_operation import TriggerOperation

if TYPE_CHECKING:
    from .device import Device


class Trigger:
    """
    A handle for a trigger with this number on the device.
    Triggers allow setting up actions that occur when a certain condition has been met or an event has occurred.
    Please note that the Triggers API is currently an experimental feature.
    Requires at least Firmware 7.06.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that this trigger belongs to.
        """
        return self._device

    @property
    def trigger_number(self) -> int:
        """
        Number of this trigger.
        """
        return self._trigger_number

    def __init__(self, device: 'Device', trigger_number: int):
        self._device = device
        self._trigger_number = trigger_number

    def enable(
            self,
            count: int = 0
    ) -> None:
        """
        Enables the trigger.
        Once a trigger is enabled, it will fire whenever its condition transitions from false to true.
        If a trigger condition is true when a disabled trigger is enabled, the trigger will fire immediately.

        Args:
            count: Number of times the trigger will fire before disabling itself.
                If count is not specified, or 0, the trigger will fire indefinitely.
        """
        if count < 0:
            raise ValueError('Invalid value; count must be 0 or positive.')

        request = dto.TriggerEnableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            count=count,
        )
        call("trigger/enable", request)

    async def enable_async(
            self,
            count: int = 0
    ) -> None:
        """
        Enables the trigger.
        Once a trigger is enabled, it will fire whenever its condition transitions from false to true.
        If a trigger condition is true when a disabled trigger is enabled, the trigger will fire immediately.

        Args:
            count: Number of times the trigger will fire before disabling itself.
                If count is not specified, or 0, the trigger will fire indefinitely.
        """
        if count < 0:
            raise ValueError('Invalid value; count must be 0 or positive.')

        request = dto.TriggerEnableRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            count=count,
        )
        await call_async("trigger/enable", request)

    def disable(
            self
    ) -> None:
        """
        Disables the trigger.
        Once disabled, the trigger will not fire and trigger actions will not run, even if trigger conditions are met.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        call("trigger/disable", request)

    async def disable_async(
            self
    ) -> None:
        """
        Disables the trigger.
        Once disabled, the trigger will not fire and trigger actions will not run, even if trigger conditions are met.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        await call_async("trigger/disable", request)

    def get_state(
            self
    ) -> TriggerState:
        """
        Gets the state of the trigger.

        Returns:
            Complete state of the trigger.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        response = call(
            "trigger/get_state",
            request,
            TriggerState.from_binary)
        return response

    async def get_state_async(
            self
    ) -> TriggerState:
        """
        Gets the state of the trigger.

        Returns:
            Complete state of the trigger.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        response = await call_async(
            "trigger/get_state",
            request,
            TriggerState.from_binary)
        return response

    def get_enabled_state(
            self
    ) -> TriggerEnabledState:
        """
        Gets the enabled state of the trigger.

        Returns:
            Whether the trigger is enabled and the number of times it will fire.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        response = call(
            "trigger/get_enabled_state",
            request,
            TriggerEnabledState.from_binary)
        return response

    async def get_enabled_state_async(
            self
    ) -> TriggerEnabledState:
        """
        Gets the enabled state of the trigger.

        Returns:
            Whether the trigger is enabled and the number of times it will fire.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        response = await call_async(
            "trigger/get_enabled_state",
            request,
            TriggerEnabledState.from_binary)
        return response

    def fire_when(
            self,
            condition: str
    ) -> None:
        """
        Set a generic trigger condition.

        Args:
            condition: The condition to set for this trigger.
        """
        request = dto.TriggerFireWhenRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            condition=condition,
        )
        call("trigger/fire_when", request)

    async def fire_when_async(
            self,
            condition: str
    ) -> None:
        """
        Set a generic trigger condition.

        Args:
            condition: The condition to set for this trigger.
        """
        request = dto.TriggerFireWhenRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            condition=condition,
        )
        await call_async("trigger/fire_when", request)

    def fire_when_encoder_distance_travelled(
            self,
            axis: int,
            distance: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition for when an encoder position has changed by a specific distance.

        Args:
            axis: The axis to monitor for this condition.
                May be set to 0 on single-axis devices only.
            distance: The measured encoder distance between trigger fires.
            unit: Units of dist.
        """
        if distance <= 0:
            raise ValueError('Invalid value; encoder distance must be a positive value.')

        request = dto.TriggerFireWhenDistanceTravelledRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            distance=distance,
            unit=unit,
        )
        call("trigger/fire_when_encoder_distance_travelled", request)

    async def fire_when_encoder_distance_travelled_async(
            self,
            axis: int,
            distance: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition for when an encoder position has changed by a specific distance.

        Args:
            axis: The axis to monitor for this condition.
                May be set to 0 on single-axis devices only.
            distance: The measured encoder distance between trigger fires.
            unit: Units of dist.
        """
        if distance <= 0:
            raise ValueError('Invalid value; encoder distance must be a positive value.')

        request = dto.TriggerFireWhenDistanceTravelledRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            distance=distance,
            unit=unit,
        )
        await call_async("trigger/fire_when_encoder_distance_travelled", request)

    def fire_when_distance_travelled(
            self,
            axis: int,
            distance: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition for when an axis position has changed by a specific distance.

        Args:
            axis: The axis to monitor for this condition.
                May be set to 0 on single-axis devices only.
            distance: The measured distance between trigger fires.
            unit: Units of dist.
        """
        if distance <= 0:
            raise ValueError('Invalid value; distance must be a positive value.')

        request = dto.TriggerFireWhenDistanceTravelledRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            distance=distance,
            unit=unit,
        )
        call("trigger/fire_when_distance_travelled", request)

    async def fire_when_distance_travelled_async(
            self,
            axis: int,
            distance: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition for when an axis position has changed by a specific distance.

        Args:
            axis: The axis to monitor for this condition.
                May be set to 0 on single-axis devices only.
            distance: The measured distance between trigger fires.
            unit: Units of dist.
        """
        if distance <= 0:
            raise ValueError('Invalid value; distance must be a positive value.')

        request = dto.TriggerFireWhenDistanceTravelledRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            distance=distance,
            unit=unit,
        )
        await call_async("trigger/fire_when_distance_travelled", request)

    def fire_when_io(
            self,
            port_type: IoPortType,
            channel: int,
            trigger_condition: TriggerCondition,
            value: float
    ) -> None:
        """
        Set a trigger condition based on an IO channel value.

        Args:
            port_type: The type of IO channel to monitor.
            channel: The IO channel to monitor.
            trigger_condition: Comparison operator.
            value: Comparison value.
        """
        if channel <= 0:
            raise ValueError('Invalid value; channel must be a positive value.')

        request = dto.TriggerFireWhenIoRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            port_type=port_type,
            channel=channel,
            trigger_condition=trigger_condition,
            value=value,
        )
        call("trigger/fire_when_io", request)

    async def fire_when_io_async(
            self,
            port_type: IoPortType,
            channel: int,
            trigger_condition: TriggerCondition,
            value: float
    ) -> None:
        """
        Set a trigger condition based on an IO channel value.

        Args:
            port_type: The type of IO channel to monitor.
            channel: The IO channel to monitor.
            trigger_condition: Comparison operator.
            value: Comparison value.
        """
        if channel <= 0:
            raise ValueError('Invalid value; channel must be a positive value.')

        request = dto.TriggerFireWhenIoRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            port_type=port_type,
            channel=channel,
            trigger_condition=trigger_condition,
            value=value,
        )
        await call_async("trigger/fire_when_io", request)

    def fire_when_setting(
            self,
            axis: int,
            setting: str,
            trigger_condition: TriggerCondition,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition based on a setting value.

        Args:
            axis: The axis to monitor for this condition.
                Set to 0 for device-scope settings.
            setting: The setting to monitor.
            trigger_condition: Comparison operator.
            value: Comparison value.
            unit: Units of value.
        """
        request = dto.TriggerFireWhenSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            setting=setting,
            trigger_condition=trigger_condition,
            value=value,
            unit=unit,
        )
        call("trigger/fire_when_setting", request)

    async def fire_when_setting_async(
            self,
            axis: int,
            setting: str,
            trigger_condition: TriggerCondition,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition based on a setting value.

        Args:
            axis: The axis to monitor for this condition.
                Set to 0 for device-scope settings.
            setting: The setting to monitor.
            trigger_condition: Comparison operator.
            value: Comparison value.
            unit: Units of value.
        """
        request = dto.TriggerFireWhenSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            setting=setting,
            trigger_condition=trigger_condition,
            value=value,
            unit=unit,
        )
        await call_async("trigger/fire_when_setting", request)

    def fire_when_absolute_setting(
            self,
            axis: int,
            setting: str,
            trigger_condition: TriggerCondition,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition based on an absolute setting value.

        Args:
            axis: The axis to monitor for this condition.
                Set to 0 for device-scope settings.
            setting: The setting to monitor.
            trigger_condition: Comparison operator.
            value: Comparison value.
            unit: Units of value.
        """
        request = dto.TriggerFireWhenSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            setting=setting,
            trigger_condition=trigger_condition,
            value=value,
            unit=unit,
        )
        call("trigger/fire_when_setting_absolute", request)

    async def fire_when_absolute_setting_async(
            self,
            axis: int,
            setting: str,
            trigger_condition: TriggerCondition,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition based on an absolute setting value.

        Args:
            axis: The axis to monitor for this condition.
                Set to 0 for device-scope settings.
            setting: The setting to monitor.
            trigger_condition: Comparison operator.
            value: Comparison value.
            unit: Units of value.
        """
        request = dto.TriggerFireWhenSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            axis=axis,
            setting=setting,
            trigger_condition=trigger_condition,
            value=value,
            unit=unit,
        )
        await call_async("trigger/fire_when_setting_absolute", request)

    def fire_at_interval(
            self,
            interval: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition based on a time interval.

        Args:
            interval: The time interval between trigger fires.
            unit: Units of time.
        """
        if interval <= 0:
            raise ValueError('Invalid value; interval must be a positive value.')

        request = dto.TriggerFireAtIntervalRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            interval=interval,
            unit=unit,
        )
        call("trigger/fire_at_interval", request)

    async def fire_at_interval_async(
            self,
            interval: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Set a trigger condition based on a time interval.

        Args:
            interval: The time interval between trigger fires.
            unit: Units of time.
        """
        if interval <= 0:
            raise ValueError('Invalid value; interval must be a positive value.')

        request = dto.TriggerFireAtIntervalRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            interval=interval,
            unit=unit,
        )
        await call_async("trigger/fire_at_interval", request)

    def on_fire(
            self,
            action: TriggerAction,
            axis: int,
            command: str
    ) -> None:
        """
        Set a command to be a trigger action.

        Args:
            action: The action number to assign the command to.
            axis: The axis to on which to run this command.
                Set to 0 for device-scope settings or to run command on all axes.
            command: The command to run when the action is triggered.
        """
        request = dto.TriggerOnFireRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
            axis=axis,
            command=command,
        )
        call("trigger/on_fire", request)

    async def on_fire_async(
            self,
            action: TriggerAction,
            axis: int,
            command: str
    ) -> None:
        """
        Set a command to be a trigger action.

        Args:
            action: The action number to assign the command to.
            axis: The axis to on which to run this command.
                Set to 0 for device-scope settings or to run command on all axes.
            command: The command to run when the action is triggered.
        """
        request = dto.TriggerOnFireRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
            axis=axis,
            command=command,
        )
        await call_async("trigger/on_fire", request)

    def on_fire_set(
            self,
            action: TriggerAction,
            axis: int,
            setting: str,
            operation: TriggerOperation,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Set a trigger action to update a setting.

        Args:
            action: The action number to assign the command to.
            axis: The axis on which to change the setting.
                Set to 0 to change the setting for the device.
            setting: The name of the setting to change.
            operation: The operation to apply to the setting.
            value: Operation value.
            unit: Units of value.
        """
        request = dto.TriggerOnFireSetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
            axis=axis,
            setting=setting,
            operation=operation,
            value=value,
            unit=unit,
        )
        call("trigger/on_fire_set", request)

    async def on_fire_set_async(
            self,
            action: TriggerAction,
            axis: int,
            setting: str,
            operation: TriggerOperation,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Set a trigger action to update a setting.

        Args:
            action: The action number to assign the command to.
            axis: The axis on which to change the setting.
                Set to 0 to change the setting for the device.
            setting: The name of the setting to change.
            operation: The operation to apply to the setting.
            value: Operation value.
            unit: Units of value.
        """
        request = dto.TriggerOnFireSetRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
            axis=axis,
            setting=setting,
            operation=operation,
            value=value,
            unit=unit,
        )
        await call_async("trigger/on_fire_set", request)

    def on_fire_set_to_setting(
            self,
            action: TriggerAction,
            axis: int,
            setting: str,
            operation: TriggerOperation,
            from_axis: int,
            from_setting: str
    ) -> None:
        """
        Set a trigger action to update a setting with the value of another setting.

        Args:
            action: The action number to assign the command to.
            axis: The axis on which to change the setting.
                Set to 0 to change the setting for the device.
            setting: The name of the setting to change.
                Must have either integer or boolean type.
            operation: The operation to apply to the setting.
            from_axis: The axis from which to read the setting.
                Set to 0 to read the setting from the device.
            from_setting: The name of the setting to read.
                Must have either integer or boolean type.
        """
        request = dto.TriggerOnFireSetToSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
            axis=axis,
            setting=setting,
            operation=operation,
            from_axis=from_axis,
            from_setting=from_setting,
        )
        call("trigger/on_fire_set_to_setting", request)

    async def on_fire_set_to_setting_async(
            self,
            action: TriggerAction,
            axis: int,
            setting: str,
            operation: TriggerOperation,
            from_axis: int,
            from_setting: str
    ) -> None:
        """
        Set a trigger action to update a setting with the value of another setting.

        Args:
            action: The action number to assign the command to.
            axis: The axis on which to change the setting.
                Set to 0 to change the setting for the device.
            setting: The name of the setting to change.
                Must have either integer or boolean type.
            operation: The operation to apply to the setting.
            from_axis: The axis from which to read the setting.
                Set to 0 to read the setting from the device.
            from_setting: The name of the setting to read.
                Must have either integer or boolean type.
        """
        request = dto.TriggerOnFireSetToSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
            axis=axis,
            setting=setting,
            operation=operation,
            from_axis=from_axis,
            from_setting=from_setting,
        )
        await call_async("trigger/on_fire_set_to_setting", request)

    def clear_action(
            self,
            action: TriggerAction = TriggerAction.ALL
    ) -> None:
        """
        Clear a trigger action.

        Args:
            action: The action number to clear.
                The default option is to clear all actions.
        """
        request = dto.TriggerClearActionRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
        )
        call("trigger/clear_action", request)

    async def clear_action_async(
            self,
            action: TriggerAction = TriggerAction.ALL
    ) -> None:
        """
        Clear a trigger action.

        Args:
            action: The action number to clear.
                The default option is to clear all actions.
        """
        request = dto.TriggerClearActionRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            action=action,
        )
        await call_async("trigger/clear_action", request)

    def get_label(
            self
    ) -> str:
        """
        Returns the label for the trigger.

        Returns:
            The label for the trigger.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        response = call(
            "trigger/get_label",
            request,
            dto.StringResponse.from_binary)
        return response.value

    async def get_label_async(
            self
    ) -> str:
        """
        Returns the label for the trigger.

        Returns:
            The label for the trigger.
        """
        request = dto.TriggerEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
        )
        response = await call_async(
            "trigger/get_label",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def set_label(
            self,
            label: Optional[str]
    ) -> None:
        """
        Sets the label for the trigger.

        Args:
            label: The label to set for this trigger.
                If no value or an empty string is provided, this label is deleted.
        """
        request = dto.TriggerSetLabelRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            label=label,
        )
        call("trigger/set_label", request)

    async def set_label_async(
            self,
            label: Optional[str]
    ) -> None:
        """
        Sets the label for the trigger.

        Args:
            label: The label to set for this trigger.
                If no value or an empty string is provided, this label is deleted.
        """
        request = dto.TriggerSetLabelRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            trigger_number=self.trigger_number,
            label=label,
        )
        await call_async("trigger/set_label", request)



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from .errors import Errors


@dataclass
class DisconnectedEvent:
    """
    Event that is sent when a connection is lost.
    """

    interface_id: int = 0
    """
    The id of the interface that was disconnected.
    """

    error_type: Errors = next(first for first in Errors)
    """
    The type of error that caused the disconnection.
    """

    error_message: str = ""
    """
    The message describing the error.
    """

    @staticmethod
    def zero_values() -> 'DisconnectedEvent':
        return DisconnectedEvent(
            interface_id=0,
            error_type=next(first for first in Errors),
            error_message="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'DisconnectedEvent':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return DisconnectedEvent.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'errorType': self.error_type.value,
            'errorMessage': str(self.error_message or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'DisconnectedEvent':
        return DisconnectedEvent(
            interface_id=data.get('interfaceId'),  # type: ignore
            error_type=Errors(data.get('errorType')),  # type: ignore
            error_message=data.get('errorMessage'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "DisconnectedEvent" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "DisconnectedEvent" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "DisconnectedEvent" is not integer value.')

        if self.error_type is None:
            raise ValueError(f'Property "ErrorType" of "DisconnectedEvent" is None.')

        if not isinstance(self.error_type, Errors):
            raise ValueError(f'Property "ErrorType" of "DisconnectedEvent" is not an instance of "Errors".')

        if self.error_message is not None:
            if not isinstance(self.error_message, str):
                raise ValueError(f'Property "ErrorMessage" of "DisconnectedEvent" is not a string.')



# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .units import UnitsAndLiterals as UnitsAndLiterals, Units as Units
from .convert_exception import convert_exception as convert_exception
from .version import __version__ as __version__
from .async_utils import wait_all as wait_all
from .library import Library as Library
from .tools import Tools as Tools
from .unit_table import UnitTable as UnitTable
from .dto.axis_address import AxisAddress as AxisAddress
from .dto.channel_address import ChannelAddress as ChannelAddress
from .dto.device_db_source_type import DeviceDbSourceType as DeviceDbSourceType
from .dto.firmware_version import FirmwareVersion as FirmwareVersion
from .dto.log_output_mode import LogOutputMode as LogOutputMode
from .dto.measurement import Measurement as Measurement
from .dto.named_parameter import NamedParameter as NamedParameter
from .dto.rotation_direction import RotationDirection as RotationDirection
from .exceptions.bad_command_exception import BadCommandException as BadCommandException
from .exceptions.bad_data_exception import BadDataException as BadDataException
from .exceptions.binary_command_failed_exception import BinaryCommandFailedException as BinaryCommandFailedException
from .exceptions.command_failed_exception import CommandFailedException as CommandFailedException
from .exceptions.command_preempted_exception import CommandPreemptedException as CommandPreemptedException
from .exceptions.command_too_long_exception import CommandTooLongException as CommandTooLongException
from .exceptions.connection_closed_exception import ConnectionClosedException as ConnectionClosedException
from .exceptions.connection_failed_exception import ConnectionFailedException as ConnectionFailedException
from .exceptions.conversion_failed_exception import ConversionFailedException as ConversionFailedException
from .exceptions.device_address_conflict_exception import DeviceAddressConflictException as DeviceAddressConflictException
from .exceptions.device_busy_exception import DeviceBusyException as DeviceBusyException
from .exceptions.device_db_failed_exception import DeviceDbFailedException as DeviceDbFailedException
from .exceptions.device_detection_failed_exception import DeviceDetectionFailedException as DeviceDetectionFailedException
from .exceptions.device_failed_exception import DeviceFailedException as DeviceFailedException
from .exceptions.device_not_identified_exception import DeviceNotIdentifiedException as DeviceNotIdentifiedException
from .exceptions.driver_disabled_exception import DriverDisabledException as DriverDisabledException
from .exceptions.g_code_execution_exception import GCodeExecutionException as GCodeExecutionException
from .exceptions.g_code_syntax_exception import GCodeSyntaxException as GCodeSyntaxException
from .exceptions.incompatible_shared_library_exception import IncompatibleSharedLibraryException as IncompatibleSharedLibraryException
from .exceptions.internal_error_exception import InternalErrorException as InternalErrorException
from .exceptions.invalid_argument_exception import InvalidArgumentException as InvalidArgumentException
from .exceptions.invalid_data_exception import InvalidDataException as InvalidDataException
from .exceptions.invalid_operation_exception import InvalidOperationException as InvalidOperationException
from .exceptions.invalid_packet_exception import InvalidPacketException as InvalidPacketException
from .exceptions.invalid_park_state_exception import InvalidParkStateException as InvalidParkStateException
from .exceptions.invalid_request_data_exception import InvalidRequestDataException as InvalidRequestDataException
from .exceptions.invalid_response_exception import InvalidResponseException as InvalidResponseException
from .exceptions.io_channel_out_of_range_exception import IoChannelOutOfRangeException as IoChannelOutOfRangeException
from .exceptions.io_failed_exception import IoFailedException as IoFailedException
from .exceptions.lockstep_enabled_exception import LockstepEnabledException as LockstepEnabledException
from .exceptions.lockstep_not_enabled_exception import LockstepNotEnabledException as LockstepNotEnabledException
from .exceptions.motion_lib_exception import MotionLibException as MotionLibException
from .exceptions.movement_failed_exception import MovementFailedException as MovementFailedException
from .exceptions.movement_interrupted_exception import MovementInterruptedException as MovementInterruptedException
from .exceptions.no_device_found_exception import NoDeviceFoundException as NoDeviceFoundException
from .exceptions.no_value_for_key_exception import NoValueForKeyException as NoValueForKeyException
from .exceptions.not_supported_exception import NotSupportedException as NotSupportedException
from .exceptions.operation_failed_exception import OperationFailedException as OperationFailedException
from .exceptions.os_failed_exception import OsFailedException as OsFailedException
from .exceptions.out_of_request_ids_exception import OutOfRequestIdsException as OutOfRequestIdsException
from .exceptions.pvt_discontinuity_exception import PvtDiscontinuityException as PvtDiscontinuityException
from .exceptions.pvt_execution_exception import PvtExecutionException as PvtExecutionException
from .exceptions.pvt_mode_exception import PvtModeException as PvtModeException
from .exceptions.pvt_movement_failed_exception import PvtMovementFailedException as PvtMovementFailedException
from .exceptions.pvt_movement_interrupted_exception import PvtMovementInterruptedException as PvtMovementInterruptedException
from .exceptions.pvt_setup_failed_exception import PvtSetupFailedException as PvtSetupFailedException
from .exceptions.remote_mode_exception import RemoteModeException as RemoteModeException
from .exceptions.request_timeout_exception import RequestTimeoutException as RequestTimeoutException
from .exceptions.serial_port_busy_exception import SerialPortBusyException as SerialPortBusyException
from .exceptions.set_device_state_failed_exception import SetDeviceStateFailedException as SetDeviceStateFailedException
from .exceptions.set_peripheral_state_failed_exception import SetPeripheralStateFailedException as SetPeripheralStateFailedException
from .exceptions.setting_not_found_exception import SettingNotFoundException as SettingNotFoundException
from .exceptions.stream_discontinuity_exception import StreamDiscontinuityException as StreamDiscontinuityException
from .exceptions.stream_execution_exception import StreamExecutionException as StreamExecutionException
from .exceptions.stream_mode_exception import StreamModeException as StreamModeException
from .exceptions.stream_movement_failed_exception import StreamMovementFailedException as StreamMovementFailedException
from .exceptions.stream_movement_interrupted_exception import StreamMovementInterruptedException as StreamMovementInterruptedException
from .exceptions.stream_setup_failed_exception import StreamSetupFailedException as StreamSetupFailedException
from .exceptions.timeout_exception import TimeoutException as TimeoutException
from .exceptions.transport_already_used_exception import TransportAlreadyUsedException as TransportAlreadyUsedException
from .exceptions.unknown_request_exception import UnknownRequestException as UnknownRequestException
from .dto.exceptions.binary_command_failed_exception_data import BinaryCommandFailedExceptionData as BinaryCommandFailedExceptionData
from .dto.exceptions.command_failed_exception_data import CommandFailedExceptionData as CommandFailedExceptionData
from .dto.exceptions.command_too_long_exception_data import CommandTooLongExceptionData as CommandTooLongExceptionData
from .dto.exceptions.device_address_conflict_exception_data import DeviceAddressConflictExceptionData as DeviceAddressConflictExceptionData
from .dto.exceptions.device_db_failed_exception_data import DeviceDbFailedExceptionData as DeviceDbFailedExceptionData
from .dto.exceptions.g_code_execution_exception_data import GCodeExecutionExceptionData as GCodeExecutionExceptionData
from .dto.exceptions.g_code_syntax_exception_data import GCodeSyntaxExceptionData as GCodeSyntaxExceptionData
from .dto.exceptions.invalid_packet_exception_data import InvalidPacketExceptionData as InvalidPacketExceptionData
from .dto.exceptions.invalid_pvt_point import InvalidPvtPoint as InvalidPvtPoint
from .dto.exceptions.invalid_response_exception_data import InvalidResponseExceptionData as InvalidResponseExceptionData
from .dto.exceptions.movement_failed_exception_data import MovementFailedExceptionData as MovementFailedExceptionData
from .dto.exceptions.movement_interrupted_exception_data import MovementInterruptedExceptionData as MovementInterruptedExceptionData
from .dto.exceptions.operation_failed_exception_data import OperationFailedExceptionData as OperationFailedExceptionData
from .dto.exceptions.pvt_execution_exception_data import PvtExecutionExceptionData as PvtExecutionExceptionData
from .dto.exceptions.pvt_movement_failed_exception_data import PvtMovementFailedExceptionData as PvtMovementFailedExceptionData
from .dto.exceptions.pvt_movement_interrupted_exception_data import PvtMovementInterruptedExceptionData as PvtMovementInterruptedExceptionData
from .dto.exceptions.set_device_state_exception_data import SetDeviceStateExceptionData as SetDeviceStateExceptionData
from .dto.exceptions.set_peripheral_state_exception_data import SetPeripheralStateExceptionData as SetPeripheralStateExceptionData
from .dto.exceptions.stream_execution_exception_data import StreamExecutionExceptionData as StreamExecutionExceptionData
from .dto.exceptions.stream_movement_failed_exception_data import StreamMovementFailedExceptionData as StreamMovementFailedExceptionData
from .dto.exceptions.stream_movement_interrupted_exception_data import StreamMovementInterruptedExceptionData as StreamMovementInterruptedExceptionData



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .process_controller_mode import ProcessControllerMode as ProcessControllerMode
from .process_controller_source import ProcessControllerSource as ProcessControllerSource
from .process_controller_source_sensor import ProcessControllerSourceSensor as ProcessControllerSourceSensor



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class TriggerEnabledState:
    """
    The enabled state of a single trigger.
    Returns whether the given trigger is enabled and the number of times it will fire.
    This is a subset of the complete state, and is faster to query.
    """

    enabled: bool
    """
    The enabled state for a trigger.
    """

    fires_remaining: int
    """
    The number of remaining fires for this trigger.
    A value of -1 indicates unlimited fires remaining.
    """

    @staticmethod
    def zero_values() -> 'TriggerEnabledState':
        return TriggerEnabledState(
            enabled=False,
            fires_remaining=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TriggerEnabledState':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TriggerEnabledState.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'enabled': bool(self.enabled),
            'firesRemaining': int(self.fires_remaining),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TriggerEnabledState':
        return TriggerEnabledState(
            enabled=data.get('enabled'),  # type: ignore
            fires_remaining=data.get('firesRemaining'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.fires_remaining is None:
            raise ValueError(f'Property "FiresRemaining" of "TriggerEnabledState" is None.')

        if not isinstance(self.fires_remaining, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "FiresRemaining" of "TriggerEnabledState" is not a number.')

        if int(self.fires_remaining) != self.fires_remaining:
            raise ValueError(f'Property "FiresRemaining" of "TriggerEnabledState" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from ..named_parameter import NamedParameter


@dataclass
class AutofocusSetObjectiveParamsRequest:

    provider_id: int = 0

    interface_id: int = 0

    focus_address: int = 0

    focus_axis: int = 0

    turret_address: int = 0

    objective: int = 0

    parameters: List[NamedParameter] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'AutofocusSetObjectiveParamsRequest':
        return AutofocusSetObjectiveParamsRequest(
            provider_id=0,
            interface_id=0,
            focus_address=0,
            focus_axis=0,
            turret_address=0,
            objective=0,
            parameters=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AutofocusSetObjectiveParamsRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AutofocusSetObjectiveParamsRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'providerId': int(self.provider_id),
            'interfaceId': int(self.interface_id),
            'focusAddress': int(self.focus_address),
            'focusAxis': int(self.focus_axis),
            'turretAddress': int(self.turret_address),
            'objective': int(self.objective),
            'parameters': [item.to_dict() for item in self.parameters] if self.parameters is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AutofocusSetObjectiveParamsRequest':
        return AutofocusSetObjectiveParamsRequest(
            provider_id=data.get('providerId'),  # type: ignore
            interface_id=data.get('interfaceId'),  # type: ignore
            focus_address=data.get('focusAddress'),  # type: ignore
            focus_axis=data.get('focusAxis'),  # type: ignore
            turret_address=data.get('turretAddress'),  # type: ignore
            objective=data.get('objective'),  # type: ignore
            parameters=[NamedParameter.from_dict(item) for item in data.get('parameters')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.provider_id is None:
            raise ValueError(f'Property "ProviderId" of "AutofocusSetObjectiveParamsRequest" is None.')

        if not isinstance(self.provider_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "ProviderId" of "AutofocusSetObjectiveParamsRequest" is not a number.')

        if int(self.provider_id) != self.provider_id:
            raise ValueError(f'Property "ProviderId" of "AutofocusSetObjectiveParamsRequest" is not integer value.')

        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "AutofocusSetObjectiveParamsRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "AutofocusSetObjectiveParamsRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "AutofocusSetObjectiveParamsRequest" is not integer value.')

        if self.focus_address is None:
            raise ValueError(f'Property "FocusAddress" of "AutofocusSetObjectiveParamsRequest" is None.')

        if not isinstance(self.focus_address, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "FocusAddress" of "AutofocusSetObjectiveParamsRequest" is not a number.')

        if int(self.focus_address) != self.focus_address:
            raise ValueError(f'Property "FocusAddress" of "AutofocusSetObjectiveParamsRequest" is not integer value.')

        if self.focus_axis is None:
            raise ValueError(f'Property "FocusAxis" of "AutofocusSetObjectiveParamsRequest" is None.')

        if not isinstance(self.focus_axis, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "FocusAxis" of "AutofocusSetObjectiveParamsRequest" is not a number.')

        if int(self.focus_axis) != self.focus_axis:
            raise ValueError(f'Property "FocusAxis" of "AutofocusSetObjectiveParamsRequest" is not integer value.')

        if self.turret_address is None:
            raise ValueError(f'Property "TurretAddress" of "AutofocusSetObjectiveParamsRequest" is None.')

        if not isinstance(self.turret_address, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "TurretAddress" of "AutofocusSetObjectiveParamsRequest" is not a number.')

        if int(self.turret_address) != self.turret_address:
            raise ValueError(f'Property "TurretAddress" of "AutofocusSetObjectiveParamsRequest" is not integer value.')

        if self.objective is None:
            raise ValueError(f'Property "Objective" of "AutofocusSetObjectiveParamsRequest" is None.')

        if not isinstance(self.objective, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Objective" of "AutofocusSetObjectiveParamsRequest" is not a number.')

        if int(self.objective) != self.objective:
            raise ValueError(f'Property "Objective" of "AutofocusSetObjectiveParamsRequest" is not integer value.')

        if self.parameters is not None:
            if not isinstance(self.parameters, Iterable):
                raise ValueError('Property "Parameters" of "AutofocusSetObjectiveParamsRequest" is not iterable.')

            for i, parameters_item in enumerate(self.parameters):
                if parameters_item is None:
                    raise ValueError(f'Item {i} in property "Parameters" of "AutofocusSetObjectiveParamsRequest" is None.')

                if not isinstance(parameters_item, NamedParameter):
                    raise ValueError(f'Item {i} in property "Parameters" of "AutofocusSetObjectiveParamsRequest" is not an instance of "NamedParameter".')

                parameters_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from .oscilloscope_data_source import OscilloscopeDataSource
from .io_port_type import IoPortType


@dataclass
class OscilloscopeCaptureProperties:
    """
    The public properties of one channel of recorded oscilloscope data.
    """

    data_source: OscilloscopeDataSource
    """
    Indicates whether the data came from a setting or an I/O pin.
    """

    setting: str
    """
    The name of the recorded setting.
    """

    axis_number: int
    """
    The number of the axis the data was recorded from, or 0 for the controller.
    """

    io_type: IoPortType
    """
    Which kind of I/O port data was recorded from.
    """

    io_channel: int
    """
    Which I/O pin within the port was recorded.
    """

    @staticmethod
    def zero_values() -> 'OscilloscopeCaptureProperties':
        return OscilloscopeCaptureProperties(
            data_source=next(first for first in OscilloscopeDataSource),
            setting="",
            axis_number=0,
            io_type=next(first for first in IoPortType),
            io_channel=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'OscilloscopeCaptureProperties':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return OscilloscopeCaptureProperties.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'dataSource': self.data_source.value,
            'setting': str(self.setting or ''),
            'axisNumber': int(self.axis_number),
            'ioType': self.io_type.value,
            'ioChannel': int(self.io_channel),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'OscilloscopeCaptureProperties':
        return OscilloscopeCaptureProperties(
            data_source=OscilloscopeDataSource(data.get('dataSource')),  # type: ignore
            setting=data.get('setting'),  # type: ignore
            axis_number=data.get('axisNumber'),  # type: ignore
            io_type=IoPortType(data.get('ioType')),  # type: ignore
            io_channel=data.get('ioChannel'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.data_source is None:
            raise ValueError(f'Property "DataSource" of "OscilloscopeCaptureProperties" is None.')

        if not isinstance(self.data_source, OscilloscopeDataSource):
            raise ValueError(f'Property "DataSource" of "OscilloscopeCaptureProperties" is not an instance of "OscilloscopeDataSource".')

        if self.setting is not None:
            if not isinstance(self.setting, str):
                raise ValueError(f'Property "Setting" of "OscilloscopeCaptureProperties" is not a string.')

        if self.axis_number is None:
            raise ValueError(f'Property "AxisNumber" of "OscilloscopeCaptureProperties" is None.')

        if not isinstance(self.axis_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisNumber" of "OscilloscopeCaptureProperties" is not a number.')

        if int(self.axis_number) != self.axis_number:
            raise ValueError(f'Property "AxisNumber" of "OscilloscopeCaptureProperties" is not integer value.')

        if self.io_type is None:
            raise ValueError(f'Property "IoType" of "OscilloscopeCaptureProperties" is None.')

        if not isinstance(self.io_type, IoPortType):
            raise ValueError(f'Property "IoType" of "OscilloscopeCaptureProperties" is not an instance of "IoPortType".')

        if self.io_channel is None:
            raise ValueError(f'Property "IoChannel" of "OscilloscopeCaptureProperties" is None.')

        if not isinstance(self.io_channel, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "IoChannel" of "OscilloscopeCaptureProperties" is not a number.')

        if int(self.io_channel) != self.io_channel:
            raise ValueError(f'Property "IoChannel" of "OscilloscopeCaptureProperties" is not integer value.')



# This file is generated. Do not modify by hand.
from enum import Enum


class Errors(Enum):

    BAD_COMMAND = 0
    BAD_DATA = 1
    BINARY_COMMAND_FAILED = 2
    COMMAND_FAILED = 3
    COMMAND_PREEMPTED = 4
    COMMAND_TOO_LONG = 5
    CONNECTION_CLOSED = 6
    CONNECTION_FAILED = 7
    CONVERSION_FAILED = 8
    DEVICE_ADDRESS_CONFLICT = 9
    DEVICE_BUSY = 10
    DEVICE_DB_FAILED = 11
    DEVICE_DETECTION_FAILED = 12
    DEVICE_FAILED = 13
    DEVICE_NOT_IDENTIFIED = 14
    DRIVER_DISABLED = 15
    G_CODE_EXECUTION = 16
    G_CODE_SYNTAX = 17
    INCOMPATIBLE_SHARED_LIBRARY = 18
    INTERNAL_ERROR = 19
    INVALID_ARGUMENT = 20
    INVALID_DATA = 21
    INVALID_OPERATION = 22
    INVALID_PACKET = 23
    INVALID_PARK_STATE = 24
    INVALID_REQUEST_DATA = 25
    INVALID_RESPONSE = 26
    IO_CHANNEL_OUT_OF_RANGE = 27
    IO_FAILED = 28
    LOCKSTEP_ENABLED = 29
    LOCKSTEP_NOT_ENABLED = 30
    MOVEMENT_FAILED = 31
    MOVEMENT_INTERRUPTED = 32
    NO_DEVICE_FOUND = 33
    NO_VALUE_FOR_KEY = 34
    NOT_SUPPORTED = 35
    OPERATION_FAILED = 36
    OS_FAILED = 37
    OUT_OF_REQUEST_IDS = 38
    PVT_DISCONTINUITY = 39
    PVT_EXECUTION = 40
    PVT_MODE = 41
    PVT_MOVEMENT_FAILED = 42
    PVT_MOVEMENT_INTERRUPTED = 43
    PVT_SETUP_FAILED = 44
    REMOTE_MODE = 45
    REQUEST_TIMEOUT = 46
    SERIAL_PORT_BUSY = 47
    SET_DEVICE_STATE_FAILED = 48
    SET_PERIPHERAL_STATE_FAILED = 49
    SETTING_NOT_FOUND = 50
    STREAM_DISCONTINUITY = 51
    STREAM_EXECUTION = 52
    STREAM_MODE = 53
    STREAM_MOVEMENT_FAILED = 54
    STREAM_MOVEMENT_INTERRUPTED = 55
    STREAM_SETUP_FAILED = 56
    TIMEOUT = 57
    TRANSPORT_ALREADY_USED = 58
    UNKNOWN_REQUEST = 59



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import TYPE_CHECKING, List
from ..call import call, call_async, call_sync

from ..dto import requests as dto
from ..units import UnitsAndLiterals, Units

from ..dto.ascii.get_setting import GetSetting
from ..dto.ascii.get_setting_result import GetSettingResult

if TYPE_CHECKING:
    from .device import Device


class DeviceSettings:
    """
    Class providing access to various device settings and properties.
    """

    def __init__(self, device: 'Device'):
        self._device = device

    def get(
            self,
            setting: str,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> float:
        """
        Returns any device setting or property.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.
            unit: Units of setting.

        Returns:
            Setting value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            unit=unit,
        )
        response = call(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_async(
            self,
            setting: str,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> float:
        """
        Returns any device setting or property.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.
            unit: Units of setting.

        Returns:
            Setting value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            unit=unit,
        )
        response = await call_async(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set(
            self,
            setting: str,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Sets any device setting.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.
            value: Value of the setting.
            unit: Units of setting.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            value=value,
            unit=unit,
        )
        call("device/set_setting", request)

    async def set_async(
            self,
            setting: str,
            value: float,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> None:
        """
        Sets any device setting.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.
            value: Value of the setting.
            unit: Units of setting.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            value=value,
            unit=unit,
        )
        await call_async("device/set_setting", request)

    def get_string(
            self,
            setting: str
    ) -> str:
        """
        Returns any device setting or property as a string.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.

        Returns:
            Setting value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
        )
        response = call(
            "device/get_setting_str",
            request,
            dto.StringResponse.from_binary)
        return response.value

    async def get_string_async(
            self,
            setting: str
    ) -> str:
        """
        Returns any device setting or property as a string.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.

        Returns:
            Setting value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
        )
        response = await call_async(
            "device/get_setting_str",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def set_string(
            self,
            setting: str,
            value: str
    ) -> None:
        """
        Sets any device setting as a string.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.
            value: Value of the setting.
        """
        request = dto.DeviceSetSettingStrRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            value=value,
        )
        call("device/set_setting_str", request)

    async def set_string_async(
            self,
            setting: str,
            value: str
    ) -> None:
        """
        Sets any device setting as a string.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_settings).

        Args:
            setting: Name of the setting.
            value: Value of the setting.
        """
        request = dto.DeviceSetSettingStrRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            value=value,
        )
        await call_async("device/set_setting_str", request)

    def convert_to_native_units(
            self,
            setting: str,
            value: float,
            unit: UnitsAndLiterals
    ) -> float:
        """
        Convert arbitrary setting value to Zaber native units.

        Args:
            setting: Name of the setting.
            value: Value of the setting in units specified by following argument.
            unit: Units of the value.

        Returns:
            Setting value.
        """
        request = dto.DeviceConvertSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            value=value,
            unit=unit,
        )
        response = call_sync(
            "device/convert_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def convert_from_native_units(
            self,
            setting: str,
            value: float,
            unit: UnitsAndLiterals
    ) -> float:
        """
        Convert arbitrary setting value from Zaber native units.

        Args:
            setting: Name of the setting.
            value: Value of the setting in Zaber native units.
            unit: Units to convert value to.

        Returns:
            Setting value.
        """
        request = dto.DeviceConvertSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            from_native=True,
            setting=setting,
            value=value,
            unit=unit,
        )
        response = call_sync(
            "device/convert_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_default(
            self,
            setting: str,
            unit: UnitsAndLiterals = Units.NATIVE
    ) -> float:
        """
        Returns the default value of a setting.

        Args:
            setting: Name of the setting.
            unit: Units of setting.

        Returns:
            Default setting value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
            unit=unit,
        )
        response = call_sync(
            "device/get_setting_default",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_default_string(
            self,
            setting: str
    ) -> str:
        """
        Returns the default value of a setting as a string.

        Args:
            setting: Name of the setting.

        Returns:
            Default setting value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
        )
        response = call_sync(
            "device/get_setting_default_str",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def can_convert_native_units(
            self,
            setting: str
    ) -> bool:
        """
        Indicates if given setting can be converted from and to native units.

        Args:
            setting: Name of the setting.

        Returns:
            True if unit conversion can be performed.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
        )
        response = call_sync(
            "device/can_convert_setting",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def get_from_all_axes(
            self,
            setting: str
    ) -> List[float]:
        """
        Gets the value of an axis scope setting for each axis on the device.
        Values may be NaN where the setting is not applicable.

        Args:
            setting: Name of the setting.

        Returns:
            The setting values on each axis.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
        )
        response = call(
            "device/get_setting_from_all_axes",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    async def get_from_all_axes_async(
            self,
            setting: str
    ) -> List[float]:
        """
        Gets the value of an axis scope setting for each axis on the device.
        Values may be NaN where the setting is not applicable.

        Args:
            setting: Name of the setting.

        Returns:
            The setting values on each axis.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            setting=setting,
        )
        response = await call_async(
            "device/get_setting_from_all_axes",
            request,
            dto.DoubleArrayResponse.from_binary)
        return response.values

    def get_many(
            self,
            *settings: GetSetting
    ) -> List[GetSettingResult]:
        """
        Gets many setting values in as few device requests as possible.

        Args:
            settings: The settings to read.

        Returns:
            The setting values read.
        """
        request = dto.DeviceMultiGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            settings=list(settings),
        )
        response = call(
            "device/get_many_settings",
            request,
            dto.GetSettingResults.from_binary)
        return response.results

    async def get_many_async(
            self,
            *settings: GetSetting
    ) -> List[GetSettingResult]:
        """
        Gets many setting values in as few device requests as possible.

        Args:
            settings: The settings to read.

        Returns:
            The setting values read.
        """
        request = dto.DeviceMultiGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            settings=list(settings),
        )
        response = await call_async(
            "device/get_many_settings",
            request,
            dto.GetSettingResults.from_binary)
        return response.results

    def get_synchronized(
            self,
            *settings: GetSetting
    ) -> List[GetSettingResult]:
        """
        Gets many setting values in the same tick, ensuring their values are synchronized.
        Requires at least Firmware 7.35.

        Args:
            settings: The settings to read.

        Returns:
            The setting values read.
        """
        request = dto.DeviceMultiGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            settings=list(settings),
        )
        response = call(
            "device/get_sync_settings",
            request,
            dto.GetSettingResults.from_binary)
        return response.results

    async def get_synchronized_async(
            self,
            *settings: GetSetting
    ) -> List[GetSettingResult]:
        """
        Gets many setting values in the same tick, ensuring their values are synchronized.
        Requires at least Firmware 7.35.

        Args:
            settings: The settings to read.

        Returns:
            The setting values read.
        """
        request = dto.DeviceMultiGetSettingRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            settings=list(settings),
        )
        response = await call_async(
            "device/get_sync_settings",
            request,
            dto.GetSettingResults.from_binary)
        return response.results



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson


@dataclass
class NamedParameter:
    """
    Named parameter with optional value.
    """

    name: str
    """
    Name of the parameter.
    """

    value: Optional[float] = None
    """
    Optional value of the parameter.
    """

    @staticmethod
    def zero_values() -> 'NamedParameter':
        return NamedParameter(
            name="",
            value=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'NamedParameter':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return NamedParameter.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'name': str(self.name or ''),
            'value': float(self.value) if self.value is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'NamedParameter':
        return NamedParameter(
            name=data.get('name'),  # type: ignore
            value=data.get('value'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.name is not None:
            if not isinstance(self.name, str):
                raise ValueError(f'Property "Name" of "NamedParameter" is not a string.')

        if self.value is not None:
            if not isinstance(self.value, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "Value" of "NamedParameter" is not a number.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson
from .interface_type import InterfaceType


@dataclass
class OpenInterfaceRequest:

    interface_type: InterfaceType = next(first for first in InterfaceType)

    port_name: str = ""

    baud_rate: int = 0

    host_name: str = ""

    port: int = 0

    transport: int = 0

    reject_routed_connection: bool = False

    cloud_id: str = ""

    token: str = ""

    api: str = ""

    connection_name: Optional[str] = None

    realm: Optional[str] = None

    @staticmethod
    def zero_values() -> 'OpenInterfaceRequest':
        return OpenInterfaceRequest(
            interface_type=next(first for first in InterfaceType),
            port_name="",
            baud_rate=0,
            host_name="",
            port=0,
            transport=0,
            reject_routed_connection=False,
            cloud_id="",
            connection_name=None,
            realm=None,
            token="",
            api="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'OpenInterfaceRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return OpenInterfaceRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceType': self.interface_type.value,
            'portName': str(self.port_name or ''),
            'baudRate': int(self.baud_rate),
            'hostName': str(self.host_name or ''),
            'port': int(self.port),
            'transport': int(self.transport),
            'rejectRoutedConnection': bool(self.reject_routed_connection),
            'cloudId': str(self.cloud_id or ''),
            'connectionName': str(self.connection_name) if self.connection_name is not None else None,
            'realm': str(self.realm) if self.realm is not None else None,
            'token': str(self.token or ''),
            'api': str(self.api or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'OpenInterfaceRequest':
        return OpenInterfaceRequest(
            interface_type=InterfaceType(data.get('interfaceType')),  # type: ignore
            port_name=data.get('portName'),  # type: ignore
            baud_rate=data.get('baudRate'),  # type: ignore
            host_name=data.get('hostName'),  # type: ignore
            port=data.get('port'),  # type: ignore
            transport=data.get('transport'),  # type: ignore
            reject_routed_connection=data.get('rejectRoutedConnection'),  # type: ignore
            cloud_id=data.get('cloudId'),  # type: ignore
            connection_name=data.get('connectionName'),  # type: ignore
            realm=data.get('realm'),  # type: ignore
            token=data.get('token'),  # type: ignore
            api=data.get('api'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_type is None:
            raise ValueError(f'Property "InterfaceType" of "OpenInterfaceRequest" is None.')

        if not isinstance(self.interface_type, InterfaceType):
            raise ValueError(f'Property "InterfaceType" of "OpenInterfaceRequest" is not an instance of "InterfaceType".')

        if self.port_name is not None:
            if not isinstance(self.port_name, str):
                raise ValueError(f'Property "PortName" of "OpenInterfaceRequest" is not a string.')

        if self.baud_rate is None:
            raise ValueError(f'Property "BaudRate" of "OpenInterfaceRequest" is None.')

        if not isinstance(self.baud_rate, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "BaudRate" of "OpenInterfaceRequest" is not a number.')

        if int(self.baud_rate) != self.baud_rate:
            raise ValueError(f'Property "BaudRate" of "OpenInterfaceRequest" is not integer value.')

        if self.host_name is not None:
            if not isinstance(self.host_name, str):
                raise ValueError(f'Property "HostName" of "OpenInterfaceRequest" is not a string.')

        if self.port is None:
            raise ValueError(f'Property "Port" of "OpenInterfaceRequest" is None.')

        if not isinstance(self.port, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Port" of "OpenInterfaceRequest" is not a number.')

        if int(self.port) != self.port:
            raise ValueError(f'Property "Port" of "OpenInterfaceRequest" is not integer value.')

        if self.transport is None:
            raise ValueError(f'Property "Transport" of "OpenInterfaceRequest" is None.')

        if not isinstance(self.transport, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Transport" of "OpenInterfaceRequest" is not a number.')

        if int(self.transport) != self.transport:
            raise ValueError(f'Property "Transport" of "OpenInterfaceRequest" is not integer value.')

        if self.cloud_id is not None:
            if not isinstance(self.cloud_id, str):
                raise ValueError(f'Property "CloudId" of "OpenInterfaceRequest" is not a string.')

        if self.connection_name is not None:
            if not isinstance(self.connection_name, str):
                raise ValueError(f'Property "ConnectionName" of "OpenInterfaceRequest" is not a string.')

        if self.realm is not None:
            if not isinstance(self.realm, str):
                raise ValueError(f'Property "Realm" of "OpenInterfaceRequest" is not a string.')

        if self.token is not None:
            if not isinstance(self.token, str):
                raise ValueError(f'Property "Token" of "OpenInterfaceRequest" is not a string.')

        if self.api is not None:
            if not isinstance(self.api, str):
                raise ValueError(f'Property "Api" of "OpenInterfaceRequest" is not a string.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from .axis_type import AxisType


@dataclass
class AxisIdentity:
    """
    Representation of data gathered during axis identification.
    """

    peripheral_id: int
    """
    Unique ID of the peripheral hardware.
    """

    peripheral_name: str
    """
    Name of the peripheral.
    """

    peripheral_serial_number: int
    """
    Serial number of the peripheral, or 0 when not applicable.
    """

    is_peripheral: bool
    """
    Indicates whether the axis is a peripheral or part of an integrated device.
    """

    axis_type: AxisType
    """
    Determines the type of an axis and units it accepts.
    """

    is_modified: bool
    """
    The peripheral has hardware modifications.
    """

    @staticmethod
    def zero_values() -> 'AxisIdentity':
        return AxisIdentity(
            peripheral_id=0,
            peripheral_name="",
            peripheral_serial_number=0,
            is_peripheral=False,
            axis_type=next(first for first in AxisType),
            is_modified=False,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AxisIdentity':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AxisIdentity.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'peripheralId': int(self.peripheral_id),
            'peripheralName': str(self.peripheral_name or ''),
            'peripheralSerialNumber': int(self.peripheral_serial_number),
            'isPeripheral': bool(self.is_peripheral),
            'axisType': self.axis_type.value,
            'isModified': bool(self.is_modified),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AxisIdentity':
        return AxisIdentity(
            peripheral_id=data.get('peripheralId'),  # type: ignore
            peripheral_name=data.get('peripheralName'),  # type: ignore
            peripheral_serial_number=data.get('peripheralSerialNumber'),  # type: ignore
            is_peripheral=data.get('isPeripheral'),  # type: ignore
            axis_type=AxisType(data.get('axisType')),  # type: ignore
            is_modified=data.get('isModified'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.peripheral_id is None:
            raise ValueError(f'Property "PeripheralId" of "AxisIdentity" is None.')

        if not isinstance(self.peripheral_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "PeripheralId" of "AxisIdentity" is not a number.')

        if int(self.peripheral_id) != self.peripheral_id:
            raise ValueError(f'Property "PeripheralId" of "AxisIdentity" is not integer value.')

        if self.peripheral_name is not None:
            if not isinstance(self.peripheral_name, str):
                raise ValueError(f'Property "PeripheralName" of "AxisIdentity" is not a string.')

        if self.peripheral_serial_number is None:
            raise ValueError(f'Property "PeripheralSerialNumber" of "AxisIdentity" is None.')

        if not isinstance(self.peripheral_serial_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "PeripheralSerialNumber" of "AxisIdentity" is not a number.')

        if int(self.peripheral_serial_number) != self.peripheral_serial_number:
            raise ValueError(f'Property "PeripheralSerialNumber" of "AxisIdentity" is not integer value.')

        if self.axis_type is None:
            raise ValueError(f'Property "AxisType" of "AxisIdentity" is None.')

        if not isinstance(self.axis_type, AxisType):
            raise ValueError(f'Property "AxisType" of "AxisIdentity" is not an instance of "AxisType".')



import atexit
from concurrent.futures import ThreadPoolExecutor
from typing import List, TypeVar, Generic, Type, Literal, TypedDict, cast
from ctypes import c_void_p, c_int64
import traceback
from reactivex.subject import Subject
from reactivex import Observable, operators as rxops

from .dto_object import DtoObject
from .dto import requests as dto
from .serialization import deserialize
from .bindings import c_set_event_handler, CALLBACK


TEventData = TypeVar('TEventData', bound=DtoObject)


class Event(Generic[TEventData]):
    def __init__(self, name: str, data: TEventData) -> None:
        self.name = name
        self.data = data


events = Subject[Event[DtoObject]]()

event_executor = ThreadPoolExecutor(max_workers=1)  # pylint: disable=consider-using-with


def on_shutdown() -> None:
    event_executor.shutdown()


atexit.register(on_shutdown)


def event_handler(event_data: c_void_p, _tag: c_int64) -> None:
    try:
        event_buffers = deserialize(event_data)
        event_executor.submit(process_event_catch, event_buffers)
    except RuntimeError:
        # the error appears due to race condition with python shutting down and cannot be prevented
        pass


event_handler_cb = CALLBACK(event_handler)

c_set_event_handler(0, event_handler_cb)

EventName = Literal[
    'test/event',
    'interface/unknown_response',
    'binary/interface/unknown_response',
    'interface/alert',
    'binary/interface/reply_only',
    'interface/disconnected',
]

ParsersDict = TypedDict('ParsersDict', {
    'test/event': Type[dto.TestEvent],
    'interface/unknown_response': Type[dto.UnknownResponseEventWrapper],
    'binary/interface/unknown_response': Type[dto.UnknownBinaryResponseEventWrapper],
    'interface/alert': Type[dto.AlertEventWrapper],
    'binary/interface/reply_only': Type[dto.BinaryReplyOnlyEventWrapper],
    'interface/disconnected': Type[dto.DisconnectedEvent],
})

parsers: ParsersDict = {
    'test/event': dto.TestEvent,
    'interface/unknown_response': dto.UnknownResponseEventWrapper,
    'binary/interface/unknown_response': dto.UnknownBinaryResponseEventWrapper,
    'interface/alert': dto.AlertEventWrapper,
    'binary/interface/reply_only': dto.BinaryReplyOnlyEventWrapper,
    'interface/disconnected': dto.DisconnectedEvent,
}


def process_event_catch(event_buffers: List[bytes]) -> None:
    try:
        process_event(event_buffers)
    except:  # noqa, pylint: disable=W0702
        print("Unhandled exception in event:")
        traceback.print_exc()


def process_event(event_buffers: List[bytes]) -> None:
    event = dto.GatewayEvent.from_binary(event_buffers[0])

    event_name = cast(EventName, event.event)
    parse_event_data = parsers.get(event_name)
    if parse_event_data is None:
        raise RuntimeError(f"Event not supported: {event_name}")

    event_data = parse_event_data.from_binary(event_buffers[1])
    events.on_next(Event(event_name, event_data))


def filter_events(
        name: EventName,
        data_type: Type[TEventData],
) -> Observable[TEventData]:
    assert parsers[name] is data_type

    def filter_name(event: Event[DtoObject]) -> bool:
        return event.name == name

    def map_data(event: Event[DtoObject]) -> TEventData:
        assert isinstance(event.data, data_type)
        return event.data

    return events.pipe(
        rxops.filter(filter_name),
        rxops.map(map_data))



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import TYPE_CHECKING, List, Optional
from ..call import call, call_async, call_sync

from ..units import Units, TimeUnits
from ..dto import requests as dto
from ..ascii.warnings import Warnings
from ..dto.ascii.response import Response
from ..dto.measurement import Measurement
from ..ascii.axis import Axis
from ..ascii.axis_settings import AxisSettings
from ..ascii.storage import AxisStorage
from ..dto.product.process_controller_mode import ProcessControllerMode
from ..dto.product.process_controller_source import ProcessControllerSource

if TYPE_CHECKING:
    from .process_controller import ProcessController


class Process:
    """
    Use to drive voltage for a process such as a heater, valve, Peltier device, etc.
    Requires at least Firmware 7.35.
    """

    @property
    def controller(self) -> 'ProcessController':
        """
        Controller for this process.
        """
        return self._controller

    @property
    def process_number(self) -> int:
        """
        The process number identifies the process on the controller.
        """
        return self._process_number

    @property
    def settings(self) -> AxisSettings:
        """
        Settings and properties of this process.
        """
        return self._settings

    @property
    def storage(self) -> AxisStorage:
        """
        Key-value storage of this process.
        """
        return self._storage

    @property
    def warnings(self) -> Warnings:
        """
        Warnings and faults of this process.
        """
        return self._warnings

    def __init__(self, controller: 'ProcessController', process_number: int):
        self._controller = controller
        self._process_number = process_number
        self._axis = Axis(controller.device, process_number)
        self._settings = AxisSettings(self._axis)
        self._storage = AxisStorage(self._axis)
        self._warnings = Warnings(controller.device, process_number)

    def enable(
            self,
            enabled: bool = True
    ) -> None:
        """
        Sets the enabled state of the driver.

        Args:
            enabled: If true (default) enables drive. If false disables.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=enabled,
        )
        call("process-controller/enable", request)

    async def enable_async(
            self,
            enabled: bool = True
    ) -> None:
        """
        Sets the enabled state of the driver.

        Args:
            enabled: If true (default) enables drive. If false disables.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=enabled,
        )
        await call_async("process-controller/enable", request)

    def on(
            self,
            duration: float = 0,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Turns this process on. In manual mode, this supplies voltage; in controlled mode, it starts the control loop.

        Args:
            duration: How long to leave the process on.
            unit: Units of time.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=True,
            duration=duration,
            unit=unit,
        )
        call("process-controller/on", request)

    async def on_async(
            self,
            duration: float = 0,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Turns this process on. In manual mode, this supplies voltage; in controlled mode, it starts the control loop.

        Args:
            duration: How long to leave the process on.
            unit: Units of time.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=True,
            duration=duration,
            unit=unit,
        )
        await call_async("process-controller/on", request)

    def off(
            self
    ) -> None:
        """
        Turns this process off.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=False,
        )
        call("process-controller/on", request)

    async def off_async(
            self
    ) -> None:
        """
        Turns this process off.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=False,
        )
        await call_async("process-controller/on", request)

    def set_mode(
            self,
            mode: ProcessControllerMode
    ) -> None:
        """
        Sets the control mode of this process.

        Args:
            mode: Mode to set this process to.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            setting="process.control.mode",
            value=mode.value,
        )
        call("device/set_setting", request)

    async def set_mode_async(
            self,
            mode: ProcessControllerMode
    ) -> None:
        """
        Sets the control mode of this process.

        Args:
            mode: Mode to set this process to.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            setting="process.control.mode",
            value=mode.value,
        )
        await call_async("device/set_setting", request)

    def get_mode(
            self
    ) -> ProcessControllerMode:
        """
        Gets the control mode of this process.

        Returns:
            Control mode.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            setting="process.control.mode",
        )
        response = call(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return ProcessControllerMode(response.value)

    async def get_mode_async(
            self
    ) -> ProcessControllerMode:
        """
        Gets the control mode of this process.

        Returns:
            Control mode.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            setting="process.control.mode",
        )
        response = await call_async(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return ProcessControllerMode(response.value)

    def get_source(
            self
    ) -> ProcessControllerSource:
        """
        Gets the source used to control this process.

        Returns:
            The source providing feedback for this process.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = call(
            "process_controller/get_source",
            request,
            ProcessControllerSource.from_binary)
        return response

    async def get_source_async(
            self
    ) -> ProcessControllerSource:
        """
        Gets the source used to control this process.

        Returns:
            The source providing feedback for this process.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = await call_async(
            "process_controller/get_source",
            request,
            ProcessControllerSource.from_binary)
        return response

    def set_source(
            self,
            source: ProcessControllerSource
    ) -> None:
        """
        Sets the source used to control this process.

        Args:
            source: Sets the source that should provide feedback for this process.
        """
        request = dto.SetProcessControllerSource(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            source=source,
        )
        call("process_controller/set_source", request)

    async def set_source_async(
            self,
            source: ProcessControllerSource
    ) -> None:
        """
        Sets the source used to control this process.

        Args:
            source: Sets the source that should provide feedback for this process.
        """
        request = dto.SetProcessControllerSource(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            source=source,
        )
        await call_async("process_controller/set_source", request)

    def get_input(
            self
    ) -> Measurement:
        """
        Gets the current value of the source used to control this process.

        Returns:
            The current value of this process's controlling source.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = call(
            "process_controller/get_input",
            request,
            Measurement.from_binary)
        return response

    async def get_input_async(
            self
    ) -> Measurement:
        """
        Gets the current value of the source used to control this process.

        Returns:
            The current value of this process's controlling source.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = await call_async(
            "process_controller/get_input",
            request,
            Measurement.from_binary)
        return response

    def bridge(
            self
    ) -> None:
        """
        Creates an H-bridge between this process and its neighbor. This method is only callable on axis 1 and 3.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=True,
        )
        call("process_controller/bridge", request)

    async def bridge_async(
            self
    ) -> None:
        """
        Creates an H-bridge between this process and its neighbor. This method is only callable on axis 1 and 3.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=True,
        )
        await call_async("process_controller/bridge", request)

    def unbridge(
            self
    ) -> None:
        """
        Breaks the H-bridge between this process and its neighbor, allowing them to be independently controlled.
        This method is only callable on axis 1 and 3.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=False,
        )
        call("process_controller/bridge", request)

    async def unbridge_async(
            self
    ) -> None:
        """
        Breaks the H-bridge between this process and its neighbor, allowing them to be independently controlled.
        This method is only callable on axis 1 and 3.
        """
        request = dto.ProcessOn(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            on=False,
        )
        await call_async("process_controller/bridge", request)

    def is_bridge(
            self
    ) -> bool:
        """
        Detects if the given process is in bridging mode.

        Returns:
            Whether this process is bridged with its neighbor.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = call(
            "process_controller/is_bridge",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_bridge_async(
            self
    ) -> bool:
        """
        Detects if the given process is in bridging mode.

        Returns:
            Whether this process is bridged with its neighbor.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = await call_async(
            "process_controller/is_bridge",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def generic_command(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this process' underlying axis.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = call(
            "interface/generic_command",
            request,
            Response.from_binary)
        return response

    async def generic_command_async(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this process' underlying axis.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = await call_async(
            "interface/generic_command",
            request,
            Response.from_binary)
        return response

    def generic_command_multi_response(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this process and expect multiple responses.
        Responses are returned in order of arrival.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = call(
            "interface/generic_command_multi_response",
            request,
            dto.GenericCommandResponseCollection.from_binary)
        return response.responses

    async def generic_command_multi_response_async(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this process and expect multiple responses.
        Responses are returned in order of arrival.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = await call_async(
            "interface/generic_command_multi_response",
            request,
            dto.GenericCommandResponseCollection.from_binary)
        return response.responses

    def generic_command_no_response(
            self,
            command: str
    ) -> None:
        """
        Sends a generic ASCII command to this process without expecting a response and without adding a message ID
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            command=command,
        )
        call("interface/generic_command_no_response", request)

    async def generic_command_no_response_async(
            self,
            command: str
    ) -> None:
        """
        Sends a generic ASCII command to this process without expecting a response and without adding a message ID
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            command=command,
        )
        await call_async("interface/generic_command_no_response", request)

    def get_state(
            self
    ) -> str:
        """
        Returns a serialization of the current process state that can be saved and reapplied.

        Returns:
            A serialization of the current state of the process.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = call(
            "device/get_state",
            request,
            dto.StringResponse.from_binary)
        return response.value

    async def get_state_async(
            self
    ) -> str:
        """
        Returns a serialization of the current process state that can be saved and reapplied.

        Returns:
            A serialization of the current state of the process.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
        )
        response = await call_async(
            "device/get_state",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def set_state(
            self,
            state: str
    ) -> None:
        """
        Applies a saved state to this process.

        Args:
            state: The state object to apply to this process.
        """
        request = dto.SetStateRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            state=state,
        )
        call("device/set_state", request)

    async def set_state_async(
            self,
            state: str
    ) -> None:
        """
        Applies a saved state to this process.

        Args:
            state: The state object to apply to this process.
        """
        request = dto.SetStateRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            state=state,
        )
        await call_async("device/set_state", request)

    def can_set_state(
            self,
            state: str
    ) -> Optional[str]:
        """
        Checks if a state can be applied to this process.
        This only covers exceptions that can be determined statically such as mismatches of ID or version,
        the process of applying the state can still fail when running.

        Args:
            state: The state object to check against.

        Returns:
            An explanation of why this state cannot be set to this process.
        """
        request = dto.CanSetStateRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            state=state,
        )
        response = call(
            "device/can_set_axis_state",
            request,
            dto.CanSetStateAxisResponse.from_binary)
        return response.error

    async def can_set_state_async(
            self,
            state: str
    ) -> Optional[str]:
        """
        Checks if a state can be applied to this process.
        This only covers exceptions that can be determined statically such as mismatches of ID or version,
        the process of applying the state can still fail when running.

        Args:
            state: The state object to check against.

        Returns:
            An explanation of why this state cannot be set to this process.
        """
        request = dto.CanSetStateRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            state=state,
        )
        response = await call_async(
            "device/can_set_axis_state",
            request,
            dto.CanSetStateAxisResponse.from_binary)
        return response.error

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the process.

        Returns:
            A string that represents the process.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.controller.device.connection.interface_id,
            device=self.controller.device.device_address,
            axis=self.process_number,
            type_override="Process",
        )
        response = call_sync(
            "device/axis_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .motion_lib_exception import MotionLibException


class PvtSetupFailedException(MotionLibException):
    """
    Thrown when setting up a PVT sequence fails.
    """



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson


@dataclass
class ForgetDevicesRequest:

    interface_id: int = 0

    except_devices: List[int] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'ForgetDevicesRequest':
        return ForgetDevicesRequest(
            interface_id=0,
            except_devices=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'ForgetDevicesRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return ForgetDevicesRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'exceptDevices': [int(item) for item in self.except_devices] if self.except_devices is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'ForgetDevicesRequest':
        return ForgetDevicesRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            except_devices=data.get('exceptDevices'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "ForgetDevicesRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "ForgetDevicesRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "ForgetDevicesRequest" is not integer value.')

        if self.except_devices is not None:
            if not isinstance(self.except_devices, Iterable):
                raise ValueError('Property "ExceptDevices" of "ForgetDevicesRequest" is not iterable.')

            for i, except_devices_item in enumerate(self.except_devices):
                if except_devices_item is None:
                    raise ValueError(f'Item {i} in property "ExceptDevices" of "ForgetDevicesRequest" is None.')

                if not isinstance(except_devices_item, (int, float, decimal.Decimal)):
                    raise ValueError(f'Item {i} in property "ExceptDevices" of "ForgetDevicesRequest" is not a number.')

                if int(except_devices_item) != except_devices_item:
                    raise ValueError(f'Item {i} in property "ExceptDevices" of "ForgetDevicesRequest" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class AxesGetSettingRequest:

    interfaces: List[int] = field(default_factory=list)

    devices: List[int] = field(default_factory=list)

    axes: List[int] = field(default_factory=list)

    unit: List[UnitsAndLiterals] = field(default_factory=list)

    setting: str = ""

    @staticmethod
    def zero_values() -> 'AxesGetSettingRequest':
        return AxesGetSettingRequest(
            interfaces=[],
            devices=[],
            axes=[],
            unit=[],
            setting="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AxesGetSettingRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AxesGetSettingRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaces': [int(item) for item in self.interfaces] if self.interfaces is not None else [],
            'devices': [int(item) for item in self.devices] if self.devices is not None else [],
            'axes': [int(item) for item in self.axes] if self.axes is not None else [],
            'unit': [units_from_literals(item).value for item in self.unit] if self.unit is not None else [],
            'setting': str(self.setting or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AxesGetSettingRequest':
        return AxesGetSettingRequest(
            interfaces=data.get('interfaces'),  # type: ignore
            devices=data.get('devices'),  # type: ignore
            axes=data.get('axes'),  # type: ignore
            unit=[Units(item) for item in data.get('unit')],  # type: ignore
            setting=data.get('setting'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interfaces is not None:
            if not isinstance(self.interfaces, Iterable):
                raise ValueError('Property "Interfaces" of "AxesGetSettingRequest" is not iterable.')

            for i, interfaces_item in enumerate(self.interfaces):
                if interfaces_item is None:
                    raise ValueError(f'Item {i} in property "Interfaces" of "AxesGetSettingRequest" is None.')

                if not isinstance(interfaces_item, (int, float, decimal.Decimal)):
                    raise ValueError(f'Item {i} in property "Interfaces" of "AxesGetSettingRequest" is not a number.')

                if int(interfaces_item) != interfaces_item:
                    raise ValueError(f'Item {i} in property "Interfaces" of "AxesGetSettingRequest" is not integer value.')

        if self.devices is not None:
            if not isinstance(self.devices, Iterable):
                raise ValueError('Property "Devices" of "AxesGetSettingRequest" is not iterable.')

            for i, devices_item in enumerate(self.devices):
                if devices_item is None:
                    raise ValueError(f'Item {i} in property "Devices" of "AxesGetSettingRequest" is None.')

                if not isinstance(devices_item, (int, float, decimal.Decimal)):
                    raise ValueError(f'Item {i} in property "Devices" of "AxesGetSettingRequest" is not a number.')

                if int(devices_item) != devices_item:
                    raise ValueError(f'Item {i} in property "Devices" of "AxesGetSettingRequest" is not integer value.')

        if self.axes is not None:
            if not isinstance(self.axes, Iterable):
                raise ValueError('Property "Axes" of "AxesGetSettingRequest" is not iterable.')

            for i, axes_item in enumerate(self.axes):
                if axes_item is None:
                    raise ValueError(f'Item {i} in property "Axes" of "AxesGetSettingRequest" is None.')

                if not isinstance(axes_item, (int, float, decimal.Decimal)):
                    raise ValueError(f'Item {i} in property "Axes" of "AxesGetSettingRequest" is not a number.')

                if int(axes_item) != axes_item:
                    raise ValueError(f'Item {i} in property "Axes" of "AxesGetSettingRequest" is not integer value.')

        if self.unit is not None:
            if not isinstance(self.unit, Iterable):
                raise ValueError('Property "Unit" of "AxesGetSettingRequest" is not iterable.')

            for i, unit_item in enumerate(self.unit):
                if unit_item is None:
                    raise ValueError(f'Item {i} in property "Unit" of "AxesGetSettingRequest" is None.')

                if not isinstance(unit_item, (Units, str)):
                    raise ValueError(f'Item {i} in property "Unit" of "AxesGetSettingRequest" is not Units.')

        if self.setting is not None:
            if not isinstance(self.setting, str):
                raise ValueError(f'Property "Setting" of "AxesGetSettingRequest" is not a string.')
